from EnteroTyper.bin.compare_sequences_types import *


def test_call_sequence_comparison():
    pass


def test_compare_reports():
    pass


def test_read_allele_report():
    pass


def test_compare_keys():
    pass


def test_compare_allele_dicts():
    pass


def test_count_mismatches():
    pass
